/*
 * ExportHTML5Plugin.java 
 *
 * Copyright (c) 2015 Emmanuel PUYBARET / eTeks <info@eteks.com>. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.eteks.sweethome3d.plugin.exporthtml5;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.URL;
import java.net.URLEncoder;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.concurrent.Callable;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.jnlp.BasicService;
import javax.jnlp.ServiceManager;
import javax.jnlp.UnavailableServiceException;
import javax.swing.JComponent;
import javax.swing.JEditorPane;
import javax.swing.JOptionPane;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import javax.swing.filechooser.FileFilter;

import com.eteks.sweethome3d.model.Home;
import com.eteks.sweethome3d.model.HomeRecorder;
import com.eteks.sweethome3d.model.InterruptedRecorderException;
import com.eteks.sweethome3d.model.RecorderException;
import com.eteks.sweethome3d.plugin.Plugin;
import com.eteks.sweethome3d.plugin.PluginAction;
import com.eteks.sweethome3d.swing.FileContentManager;
import com.eteks.sweethome3d.swing.SwingViewFactory;
import com.eteks.sweethome3d.tools.OperatingSystem;
import com.eteks.sweethome3d.viewcontroller.ContentManager;
import com.eteks.sweethome3d.viewcontroller.HomeView;
import com.eteks.sweethome3d.viewcontroller.ThreadedTaskController;

/**
 * A plug-in that generates a zip file containing an HTML page able to view the edited home with Javascript/WebGL.
 * @author Emmanuel Puybaret
 */
public class ExportHTML5Plugin extends Plugin {
  @Override
  public PluginAction [] getActions() {
    return new PluginAction [] {new ExportHTML5PluginAction("com.eteks.sweethome3d.plugin.exporthtml5.ApplicationPlugin", 
        "EXPORT_TO_HTML5", getPluginClassLoader(), true)};
  }

  protected class ExportHTML5PluginAction extends PluginAction {
    private String resourceBaseName;

    public ExportHTML5PluginAction(String resourceBaseName,  String actionPrefix, ClassLoader pluginClassLoader, boolean enabled) {
      super(resourceBaseName, actionPrefix, pluginClassLoader, enabled);
      this.resourceBaseName = resourceBaseName;
    }
  
    /**
     * Exports edited home.
     */
    public void execute() {
      final ResourceBundle resource = ResourceBundle.getBundle(this.resourceBaseName, 
          Locale.getDefault(), getPluginClassLoader());
      final HomeView homeView = getHomeController().getView();
      
      try {
        // Ignore plug-in in protected Java Web Start environment 
        ServiceManager.lookup("javax.jnlp.FileSaveService");
        // Use an uneditable editor pane to let user select text in dialog
        JEditorPane messagePane = new JEditorPane("text/html", 
            resource.getString("exportHTML5JavaWebStartInfo.message"));
        messagePane.setOpaque(false);
        messagePane.setEditable(false);
        try { 
          // Lookup the javax.jnlp.BasicService object 
          final BasicService service = (BasicService)ServiceManager.lookup("javax.jnlp.BasicService");
          // If basic service supports  web browser
          if (service.isWebBrowserSupported()) {
            // Add a listener that displays hyperlinks content in browser
            messagePane.addHyperlinkListener(new HyperlinkListener() {
              public void hyperlinkUpdate(HyperlinkEvent ev) {
                if (ev.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
                  service.showDocument(ev.getURL()); 
                }
              }
            });
          }
        } catch (UnavailableServiceException ex) {
          // Too bad : service is unavailable             
        } 
        
        String title = resource.getString("exportHTML5JavaWebStartInfo.title");
        JOptionPane.showMessageDialog((JComponent)homeView, messagePane, title, JOptionPane.WARNING_MESSAGE);
        return;
      } catch (UnavailableServiceException ex) {
      }
      
      ContentManager contentManagerWithZipExtension = new FileContentManager(getUserPreferences()) {
        private final String ZIP_EXTENSION = ".zip";
        private final FileFilter ZIP_FILE_FILTER = new FileFilter() {
          @Override
          public boolean accept(File file) {
            // Accept directories and ZIP files
            return file.isDirectory()
                || file.getName().toLowerCase().endsWith(ZIP_EXTENSION);
          }
          
          @Override
          public String getDescription() {
            return "ZIP";
          }
        };
        
        @Override
        public String getDefaultFileExtension(ContentType contentType) {
          if (contentType == ContentType.USER_DEFINED) {
            return ZIP_EXTENSION;
          } else {
            return super.getDefaultFileExtension(contentType);
          }
        }
        
        @Override
        protected String [] getFileExtensions(ContentType contentType) {
          if (contentType == ContentType.USER_DEFINED) {
            return new String [] {ZIP_EXTENSION};
          } else {
            return super.getFileExtensions(contentType);
          }
        }
        
        @Override
        protected FileFilter [] getFileFilter(ContentType contentType) {
          if (contentType == ContentType.USER_DEFINED) {
            return new FileFilter [] {ZIP_FILE_FILTER};
          } else {
            return super.getFileFilter(contentType);
          }
        }
      };
      
      // Request a file name 
      final String exportedFile = contentManagerWithZipExtension.showSaveDialog(homeView,
          resource.getString("exportHTML5Dialog.title"), 
          ContentManager.ContentType.USER_DEFINED, getHome().getName());
      if (exportedFile != null) {
        // Export to HTML5 in a threaded task
        Callable<Void> exportToObjTask = new Callable<Void>() {
          public Void call() throws RecorderException {
            exportHomeToHTML5(getHome().clone(), new File(exportedFile), 
                resource.getString("exportHTML5ResourcesBase"),
                resource.getString("exportHTML5Resources").split("\\s"));
            return null;
          }
        };
        ThreadedTaskController.ExceptionHandler exceptionHandler = 
            new ThreadedTaskController.ExceptionHandler() {
          public void handleException(Exception ex) {
            if (!(ex instanceof InterruptedRecorderException)) {
              ex.printStackTrace();
              getHomeController().getView().showError(
                  String.format(resource.getString("exportHTML5Error.message"), ex.getMessage()));
            }
          }
        };
        new ThreadedTaskController(exportToObjTask, 
            resource.getString("exportHTML5Message"), exceptionHandler, 
            getUserPreferences(), new SwingViewFactory()).executeTask(homeView);
      }
    }
    
    /**
     * Exports the given <code>home</code> to HTML5.
     * Caution: this method is called from a separate thread.
     */
    private void exportHomeToHTML5(Home home, 
                                   File exportedFile, 
                                   String includedFilesBase,
                                   String [] includedFiles) throws RecorderException {
      File homeFile = null;
      ZipOutputStream zipOut = null;
      boolean finished = false;
      try {
        // Create a temporary file 
        homeFile = OperatingSystem.createTemporaryFile("Home", ".zip");
        getHomeRecorder().writeHome(home, homeFile.getAbsolutePath());
        // Add home file to the selected file
        zipOut = new ZipOutputStream(new FileOutputStream(exportedFile));
        zipOut.setLevel(9);
        String exportedHomeFile = exportedFile.getName();
        writeZipEntry(zipOut, exportedHomeFile, homeFile.toURI().toURL(), null);
        // Add additional resources
        for (String includedFile : includedFiles) {
          writeZipEntry(zipOut, includedFile, 
              ExportHTML5Plugin.class.getResource(includedFilesBase + (includedFilesBase.endsWith("/") ? "" : "/") + includedFile), 
              includedFile.endsWith(".html") 
                  ? URLEncoder.encode(exportedHomeFile, "UTF-8").replace("+", "%20")
                  : null);
        }
        zipOut.finish();
        finished = true;
      } catch (IOException ex) {
        throw new RecorderException("Couldn't save " + exportedFile.getName(), ex);
      } finally {
        if (zipOut != null) {
          try {
            zipOut.close();
            if (!finished) {
              exportedFile.delete();
            }
          } catch (IOException ex) {
            throw new RecorderException("Couldn't close " + exportedFile.getName(), ex);
          }
        }
        if (homeFile != null) {
          homeFile.delete();
        }
      }    
    }
    
    private void writeZipEntry(ZipOutputStream zipOut, String entryName, 
                               URL source, String exportHomeFile) throws IOException {
      byte [] buffer = new byte [8192];
      InputStream in = null;
      try {
        zipOut.putNextEntry(new ZipEntry(entryName));
        in = source.openStream();
        if (exportHomeFile != null) {
          BufferedReader reader = new BufferedReader(new InputStreamReader(in, "UTF-8"));
          Writer writer = new OutputStreamWriter(zipOut, "UTF-8");
          for (String line; (line = reader.readLine()) != null; ) {
            writer.write(line.replace("${exportedHome}", exportHomeFile)
                .replace("${version}", getVersion()));
            writer.write("\n");
          }
          writer.flush();
        } else {
          for (int size; (size = in.read(buffer)) != -1; ) {
            zipOut.write(buffer, 0, size);
          }
        }      
        zipOut.closeEntry();  
      } finally {
        if (in != null) {          
          in.close();
        }
      }
    }
    
    protected HomeRecorder getHomeRecorder() {
      return new HomeXMLFileRecorder(9, 
          HomeXMLFileRecorder.INCLUDE_VIEWER_DATA 
          | HomeXMLFileRecorder.INCLUDE_HOME_STRUCTURE 
          | HomeXMLFileRecorder.REDUCE_IMAGES 
          | HomeXMLFileRecorder.CONVERT_MODELS_TO_OBJ_FORMAT, 256);
    }
  }
}
